@@ -11,7 +11,7 @@
   FOR A PARTICULAR PURPOSE
 
 ****************************************************************************/
-#include "_compat.h"
+#include "Python.h"
 
 #define BTREETEMPLATE_C "$Id$\n"
 
@@ -20,17 +20,13 @@ _get_max_size(BTree *self, PyObject *name, long default_max)
 {
   PyObject *size;
   long isize;
-
   size = PyObject_GetAttr(OBJECT(OBJECT(self)->ob_type), name);
   if (size == NULL) {
-    PyErr_Clear();
-    return default_max;
+      PyErr_Clear();
+      return default_max;
   }
-#ifdef PY3K
+
   isize = PyLong_AsLong(size);
-#else
-  isize = PyInt_AsLong(size);
-#endif
 
   Py_DECREF(size);
   if (isize <= 0 && ! PyErr_Occurred()) {
@@ -48,7 +44,7 @@ _max_internal_size(BTree *self)
   long isize;
 
   if (self->max_internal_size > 0) return self->max_internal_size;
-  isize = _get_max_size(self, max_internal_size_str, DEFAULT_MAX_BTREE_SIZE);
+  isize = _get_max_size(self, max_internal_size_str, -1);
   self->max_internal_size = isize;
   return isize;
 }
@@ -59,7 +55,7 @@ _max_leaf_size(BTree *self)
   long isize;
 
   if (self->max_leaf_size > 0) return self->max_leaf_size;
-  isize = _get_max_size(self, max_leaf_size_str, DEFAULT_MAX_BUCKET_SIZE);
+  isize = _get_max_size(self, max_leaf_size_str, -1);
   self->max_leaf_size = isize;
   return isize;
 }
@@ -271,7 +267,7 @@ _BTree_get(BTree *self, PyObject *keyarg, int has_key, int replace_type_err)
     {
         /* empty BTree */
         if (has_key)
-            result = INT_FROM_LONG(0);
+            result = PyLong_FromLong(0);
         else
             PyErr_SetObject(PyExc_KeyError, keyarg);
     }
@@ -1035,6 +1031,14 @@ BTree__p_deactivate(BTree *self, PyObject *args, PyObject *keywords)
         }
     }
 
+    /*
+      Always clear our node size cache, whether we're in a jar or not. It is
+      only read from the type anyway, and we'll do so on the next write after
+      we get activated.
+    */
+    self->max_internal_size = 0;
+    self->max_leaf_size = 0;
+
     if (self->jar && self->oid)
     {
         ghostify = self->state == cPersistent_UPTODATE_STATE;
@@ -1179,6 +1183,7 @@ _BTree_setstate(BTree *self, PyObject *state, int noval)
     PyObject *items, *firstbucket = NULL;
     BTreeItem *d;
     int len, l, i, copied=1;
+    PyTypeObject *leaftype = (noval ? &SetType : &BucketType);
 
     if (_BTree_clear(self) < 0)
         return -1;
@@ -1204,8 +1209,7 @@ _BTree_setstate(BTree *self, PyObject *state, int noval)
     }
 
     len = PyTuple_Size(items);
-    if (len < 0)
-        return -1;
+    ASSERT(len >= 0, "_BTree_setstate: items tuple has negative size", -1);
     len = (len + 1) / 2;
 
     assert(len > 0); /* If the BTree is empty, it's state is None. */
@@ -1248,6 +1252,17 @@ _BTree_setstate(BTree *self, PyObject *state, int noval)
         }
         else
         {
+            if (!(SameType_Check(self, v) ||
+                  PyObject_IsInstance(v, (PyObject *)leaftype)))
+            {
+                PyErr_Format(PyExc_TypeError,
+                             "tree child %s is neither %s nor %s",
+                             Py_TYPE(v)->tp_name,
+                             Py_TYPE(self)->tp_name,
+                             leaftype->tp_name);
+                return -1;
+            }
+
             d->child = (Sized *)v;
             Py_INCREF(v);
         }
@@ -1257,8 +1272,7 @@ _BTree_setstate(BTree *self, PyObject *state, int noval)
     if (!firstbucket)
         firstbucket = (PyObject *)self->data->child;
 
-    if (!PyObject_IsInstance(firstbucket, (PyObject *)
-                            (noval ? &SetType : &BucketType)))
+    if (!PyObject_IsInstance(firstbucket, (PyObject *)leaftype))
     {
         PyErr_SetString(PyExc_TypeError,
                         "No firstbucket in non-empty BTree");
@@ -1964,7 +1978,7 @@ BTree_getm(BTree *self, PyObject *args)
         return NULL;
     if ((r=_BTree_get(self, key, 0, _BGET_REPLACE_TYPE_ERROR)))
         return r;
-    UNLESS (PyErr_ExceptionMatches(PyExc_KeyError))
+    UNLESS (BTree_ShouldSuppressKeyError())
         return NULL;
     PyErr_Clear();
     Py_INCREF(d);
@@ -1988,7 +2002,7 @@ BTree_setdefault(BTree *self, PyObject *args)
     /* The key isn't in the tree.  If that's not due to a KeyError exception,
     * pass back the unexpected exception.
     */
-    if (! PyErr_ExceptionMatches(PyExc_KeyError))
+    if (! BTree_ShouldSuppressKeyError())
         return NULL;
     PyErr_Clear();
 
@@ -2029,7 +2043,7 @@ BTree_pop(BTree *self, PyObject *args)
     /* The key isn't in the tree.  If that's not due to a KeyError exception,
     * pass back the unexpected exception.
     */
-    if (! PyErr_ExceptionMatches(PyExc_KeyError))
+    if (! BTree_ShouldSuppressKeyError())
         return NULL;
 
     if (failobj != NULL)
@@ -2048,6 +2062,43 @@ BTree_pop(BTree *self, PyObject *args)
     return NULL;
 }
 
+
+static PyObject*
+BTree_popitem(BTree* self, PyObject* args)
+{
+    PyObject* key = NULL;
+    PyObject* pop_args = NULL;
+    PyObject* result_val = NULL;
+    PyObject* result = NULL;
+
+    if (PyTuple_Size(args) != 0) {
+        PyErr_SetString(PyExc_TypeError, "popitem(): Takes no arguments.");
+        return NULL;
+    }
+
+    key = BTree_minKey(self, args); /* reuse existing empty tuple. */
+    if (!key) {
+        PyErr_Clear();
+        PyErr_SetString(PyExc_KeyError, "popitem(): empty BTree.");
+        return NULL;
+    }
+
+    pop_args = PyTuple_Pack(1, key);
+    if (pop_args) {
+        result_val = BTree_pop(self, pop_args);
+        Py_DECREF(pop_args);
+        if (result_val) {
+            result = PyTuple_Pack(2, key, result_val);
+            Py_DECREF(result_val);
+        }
+    }
+
+    Py_DECREF(key);
+    return result;
+}
+
+
+
 /* Search BTree self for key.  This is the sq_contains slot of the
  * PySequenceMethods.
  *
@@ -2064,10 +2115,10 @@ BTree_contains(BTree *self, PyObject *key)
 
     if (asobj != NULL)
     {
-        result = INT_AS_LONG(asobj) ? 1 : 0;
+        result = PyLong_AsLong(asobj) ? 1 : 0;
         Py_DECREF(asobj);
     }
-    else if (PyErr_ExceptionMatches(PyExc_KeyError))
+    else if (BTree_ShouldSuppressKeyError())
     {
         PyErr_Clear();
         result = 0;
@@ -2101,7 +2152,7 @@ BTree_addUnique(BTree *self, PyObject *args)
 
     if ((grew=_BTree_set(self, key, v, 1, 0)) < 0)
         return NULL;
-    return INT_FROM_LONG(grew);
+    return PyLong_FromLong(grew);
 }
 
 /**************************************************************************/
@@ -2220,6 +2271,10 @@ static struct PyMethodDef BTree_methods[] = {
      "If key is not found, d is returned if given, otherwise KeyError\n"
      "is raised."},
 
+    {"popitem", (PyCFunction)BTree_popitem, METH_VARARGS,
+     "D.popitem() -> (k, v), remove and return some (key, value) pair\n"
+     "as a 2-tuple; but raise KeyError if D is empty."},
+
     {"maxKey", (PyCFunction) BTree_maxKey, METH_VARARGS,
      "maxKey([max]) -> key\n\n"
      "Return the largest key in the BTree.  If max is specified, return\n"
@@ -2425,22 +2480,89 @@ BTree_nonzero(BTree *self)
 
 static PyNumberMethods BTree_as_number_for_nonzero = {
     0,                                      /* nb_add */
-    0,                                      /* nb_subtract */
+    bucket_sub,                             /* nb_subtract */
     0,                                      /* nb_multiply */
-#ifndef PY3K
-    0,                                      /* nb_divide */
-#endif
     0,                                      /* nb_remainder */
     0,                                      /* nb_divmod */
     0,                                      /* nb_power */
     0,                                      /* nb_negative */
     0,                                      /* nb_positive */
     0,                                      /* nb_absolute */
-    (inquiry)BTree_nonzero                  /* nb_nonzero */
+    (inquiry)BTree_nonzero,                 /* nb_nonzero */
+    (unaryfunc)0,                           /* nb_invert */
+    (binaryfunc)0,                          /* nb_lshift */
+    (binaryfunc)0,                          /* nb_rshift */
+    bucket_and,                             /* nb_and */
+    (binaryfunc)0,                          /* nb_xor */
+    bucket_or,                              /* nb_or */
+};
+
+static PyObject* BTreeType_setattro_allowed_names; /* initialized in module */
+
+static int
+BTreeType_setattro(PyTypeObject* type, PyObject* name, PyObject* value)
+{
+    /*
+      type.tp_setattro prohibits setting any attributes on a built-in type,
+      so we need to use our own (metaclass) type to handle it. The set of
+      allowable values needs to be carefully controlled (e.g., setting methods
+      would be bad).
+
+      Alternately, we could use heap-allocated types when they are supported
+      an all the versions we care about, because those do allow setting attributes.
+    */
+    int allowed;
+    allowed = PySequence_Contains(BTreeType_setattro_allowed_names, name);
+    if (allowed < 0) {
+        return -1;
+    }
+
+    if (allowed) {
+        PyDict_SetItem(type->tp_dict, name, value);
+        PyType_Modified(type);
+        if (PyErr_Occurred()) {
+            return -1;
+        }
+        return 0;
+    }
+    return PyType_Type.tp_setattro((PyObject*)type, name, value);
+}
+
+static PyTypeObject BTreeTypeType = {
+    PyVarObject_HEAD_INIT(NULL, 0)
+    MODULE_NAME MOD_NAME_PREFIX "BTreeType",
+    0, /* tp_basicsize */
+    0, /* tp_itemsize */
+    0, /* tp_dealloc */
+    0, /* tp_print */
+    0, /* tp_getattr */
+    0, /* tp_setattr */
+    0, /* tp_compare */
+    0, /* tp_repr */
+    0, /* tp_as_number */
+    0, /* tp_as_sequence */
+    0, /* tp_as_mapping */
+    0, /* tp_hash */
+    0, /* tp_call */
+    0, /* tp_str */
+    0, /* tp_getattro */
+    (setattrofunc)BTreeType_setattro, /* tp_setattro */
+    0, /* tp_as_buffer */
+    Py_TPFLAGS_DEFAULT |
+    Py_TPFLAGS_BASETYPE, /* tp_flags */
+    0, /* tp_doc */
+    0, /* tp_traverse */
+    0, /* tp_clear */
+    0, /* tp_richcompare */
+    0, /* tp_weaklistoffset */
+    0, /* tp_iter */
+    0, /* tp_iternext */
+    0, /* tp_methods */
+    0, /* tp_members */
 };
 
 static PyTypeObject BTreeType = {
-    PyVarObject_HEAD_INIT(NULL, 0)
+    PyVarObject_HEAD_INIT(&BTreeTypeType, 0)
     MODULE_NAME MOD_NAME_PREFIX "BTree",    /* tp_name */
     sizeof(BTree),                          /* tp_basicsize */
     0,                                      /* tp_itemsize */
