@@ -26,18 +26,14 @@
 
 #define MOD_NAME_PREFIX "UU"
 
-#define DEFAULT_MAX_BUCKET_SIZE 120
-#define DEFAULT_MAX_BTREE_SIZE 500
+
+
 #define ZODB_UNSIGNED_KEY_INTS
 #define ZODB_UNSIGNED_VALUE_INTS
 
-#include "_compat.h"
+#include "Python.h"
 #include "intkeymacros.h"
 #include "intvaluemacros.h"
 
-#ifdef PY3K
 #define INITMODULE PyInit__UUBTree
-#else
-#define INITMODULE init_UUBTree
-#endif
 #include "BTreeModuleTemplate.c"
